package tea

import fig.basic.CharEncUtils
import scala.collection.mutable.LinkedHashMap

trait LineBasedExecutor {
  def forceLaunch = {} // Don't do things lazily
  def apply(input:String) : String
  def finish : Unit
}

// Execute command using stdin/stdout
// WARNING: make sure that command flushes, otherwise could get deadlock.
class DirectLineBasedExecutor(command:String) extends LineBasedExecutor {
  val process = Runtime.getRuntime.exec(Array("bash", "-c", command))
  val out = CharEncUtils.getWriter(process.getOutputStream)
  val in = CharEncUtils.getReader(process.getInputStream)

  def apply(input:String) : String = this.synchronized {
    out.println(input)
    out.flush
    println("SENT "+input)
    try {
      val s = in.readLine
      println("RECEIVED "+s)
      if (s == null) throw Utils.fails("Command failed: %s", command)
      s
    }
    catch { case e:java.io.IOException => throw Utils.fails("Command failed: %s", command) }
  }
  def finish = {
    in.close
    out.close
    try { process.waitFor() }
    catch { case _:java.lang.InterruptedException => }
  }
}

// Usage in this order: apply(), save(), finish()
class CachedLineBasedExecutor(command:String, cachedPath:String, delim:String="\t") extends LineBasedExecutor {
  val cache = new LinkedHashMap[String,String]
  load
  var executor : LineBasedExecutor = null

  override def forceLaunch = {
    if (executor == null)
      executor = new DirectLineBasedExecutor(command)
  }

  def apply(input:String) : String = this.synchronized {
    cache.getOrElseUpdate(input, {
      forceLaunch
      val output = executor(input)
      if (cachedPath != null) { // Save immediately
        val f = fig.basic.IOUtils.openOutAppendHard(cachedPath)
        f.println(input+delim+output)
        f.close
      }
      output
    })
  }

  def load = if (cachedPath != null) Utils.track("Loading %s", cachedPath) {
    Utils.foreachLine(cachedPath, { line:String =>
      val Array(input, output) = line.split(delim)
      cache(input) = output
      true
    })
  }

  def save = if (cachedPath != null) Utils.track("Saving %s to %s", cache.size, cachedPath) {
    Utils.writeLines(cachedPath, { puts:(String=>Any) =>
      cache.foreach { case (input, output) =>
        puts(input+delim+output)
      }
    })
  }

  def finish = if (executor != null) {
    save
    executor.finish
  }
}
