package tea

import Utils.{map,foreach,fmt,fmts,fmt1,fmt2,returnFirst,assertValid}
import Utils.fails
import Globals._

/*
Parses a LISP-like file
Features:
  Comments: # ...
  String tokens: "... \n\" ..."
  Macro substitution $foo
  Marco definition: (.define foo (... ...))

A tree is either a string or an array of trees.
In code, there are two ways of representing this:
 - Raw: use String and Array[Any] (more efficient, but type unsafe)
 - Wrapped: Tree objects
*/

trait Tree
case class BaseTree(value:String) extends Tree {
  override def toString = value
}
case class CompositeTree(children:List[Tree]) extends Tree {
  override def toString = "("+children.mkString(" ")+")"
}

class TreeStringUtils {
  val quoteChar = 34 // '"' has syntax highlighting issues

  // Can set these to null to disable
  var commentChar = '#'
  var macroChar = '$'

  val subst = new HashMap[String,Any]

  def rawToTree(x:Any) : Tree = x match {
    case x:String => BaseTree(x)
    case x:Array[Any] => CompositeTree(x.map(rawToTree).toList)
    case _ => throw fails("Bad type: %s", x)
  }
  def treeToRaw(tree:Tree) : Any = tree match {
    case BaseTree(value) => value
    case CompositeTree(children) => children.map(treeToRaw).toArray
  }

  def stringToNodes(s:String) : Array[Any] = {
    val in = new java.io.StringReader(s)
    val nodes = new ArrayBuffer[Any]
    foreachNode(in, { node:Any =>
      nodes += node
      true
    })
    in.close
    nodes.toArray
  }

  def readNodes(path:String) : Array[Any] = {
    val nodes = new ArrayBuffer[Any]
    foreachNode(path, { node:Any =>
      nodes += node
      true
    })
    nodes.toArray
  }

  // Return an array of nodes
  // A node is either a string or an array of children nodes
  def foreachNode(path:String, f:Any=>Boolean) : Unit = {
    val in = fig.basic.IOUtils.openInHard(path)
    foreachNode(in, f)
    in.close
  }

  def foreachNode(in:java.io.Reader, f:Any=>Boolean) : Unit = {
    var lineNum = 1
    var colNum = 0
    var nextc = -1 // Next character to be read
    def doRead = {
      val c = in.read
      if (c == '\n') { lineNum += 1; colNum = 0 }
      else colNum += 1
      c
    }
    def peekChar = {
      if (nextc == -1) nextc = doRead
      nextc
    }
    def readChar = {
      if (nextc == -1) doRead
      else { val c = nextc; nextc = -1; c }
    }

    // Also skip comments
    def skipAllSpace = {
      var c = peekChar
      var inComment = false
      while (c != -1 && (inComment || c == commentChar || Character.isWhitespace(c))) {
        if (c == commentChar) inComment = true
        else if (c == '\n') inComment = false
        readChar
        c = peekChar
      } 
    }

    def readString = {
      skipAllSpace
      var c = peekChar
      val buf = new StringBuilder
      if (c == quoteChar) { // Read until the end of the quote
        val savePos = (lineNum, colNum)
        var escaped = false 
        readChar
        c = readChar
        while (c != -1 && (escaped || c != quoteChar)) {
          if (!escaped && c == '\\') escaped = true
          else {
            if (escaped && c == 'n') c = '\n'
            else if (escaped && c == 't') c = '\t'
            escaped = false
            buf += c.toChar
          }
          c = readChar
        }
        if (c == -1) throw new RuntimeException("Quote at "+savePos+" not closed")
        buf.toString
      }
      else { // Read until the end of the next delimiter
        val isMacro = {
          if (c == macroChar) true
          else { buf += c.toChar; false }
        }
        readChar
        c = peekChar
        while (c != -1 && !(Character.isWhitespace(c) || c == macroChar || c == '(' || c == ')')) {
          buf += c.toChar
          readChar
          c = peekChar
        }
        val s = buf.toString
        if (isMacro) subst.getOrElse(s, throw new RuntimeException("Variable '"+s+"' is undefined"))
        else s
      }
    }

    def readNode : Any = {
      skipAllSpace
      var c = peekChar
      if (c == -1) null
      else if (c == '(') {
        val savePos = (lineNum, colNum)
        readChar // '('
        val l = new ArrayBuffer[Any]
        var done = false
        while (peekChar != ')') {
          l += readNode
          if (l.last == null) throw new RuntimeException("'(' opened at "+savePos+" not closed")
          skipAllSpace
        }
        readChar
        l.toArray match {
          case Array(".define", name:String, value:Any) => subst(name) = value; readNode
          case a:Array[Any] => a
        }
      }
      else if (c == ')')
        throw new RuntimeException("Extra ')' at "+(lineNum,colNum))
      else
        readString
    }

    def readNodes : Unit = {
      val node = readNode
      if (node != null) { if (f(node)) readNodes }
    }
    readNodes
  }

  def render(x:Any) : Array[String] = {
    val maxLen = 100
    x match {
      case x:Tree => render(treeToRaw(x))
      case null => Array("null")
      case x:String => 
        if (x == "" || x.exists { c:Char => c == quoteChar || c == '\\' || Character.isWhitespace(c) })
          Array("\""+x.replaceAll("\\\\", "\\\\\\\\").replaceAll("\n", "\\\\n").replaceAll("\t", "\\\\t")+"\"")
        else
          Array(x)
      case in:Array[Any] =>
        val out : Array[Array[String]] = in.map(render(_))
        // See if we should render it all on one line
        if (out.forall(_.size == 1) && ((0 /: out)(_ + _(0).size)) <= maxLen)
          Array("("+out.flatten.mkString(" ")+")")
        else {
          // Can't render it all on one line, so break it up into multiple lines,
          // where subsequent lines are indented.
          val lines = new ArrayBuffer[String]
          val line = new StringBuilder
          var nextDelim = ""
          def add(token:String) = {
            if (line.size + nextDelim.size + token.size > maxLen)
              flush
            line.append(nextDelim)
            line.append(token)
            nextDelim = " "
          }
          def flush = {
            if (line.size > 0) {
              lines += line.toString
              line.clear
              nextDelim = "  "
            }
          }
          line.append("(")
          out.foreach { z:Array[String] =>
            if (z.size > 1) { // If one of the children requires two lines, they need to go on their own lines
              flush 
              z.foreach(lines += nextDelim+_)
            }
            else
              add(z(0))
          }
          flush
          lines(lines.size-1) += ")"
          lines.toArray
        }
      case _ => throw new RuntimeException("Unexpected type (expect string or array): "+x)
    }
  }
}

object TreeStringUtils {
  def main(args:Array[String]) = {
    val u = new TreeStringUtils
    args.foreach { path:String =>
      u.render(u.readNodes(path)).foreach(println)
    }
  }
}
