"""Convert Geoquery logical forms to something like Lambda-DCS."""
import collections
import glob
import os
import sys

IN_DIR = os.path.join(
    os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))),
    'data/geo880/processed-lesscopy')
OUT_DIR = os.path.join(
    os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))),
    'data/geo880/processed-lambdadcs')

def read_examples(filename):
  with open(filename) as f:
    return [tuple(line.strip().split('\t')) for line in f]

ASTNode = collections.namedtuple('ASTNode', ['label', 'children'])
IDS = ['_cityid', '_countryid', '_placeid', '_riverid', '_stateid']
UNARIES = [ '_capital', '_city', '_lake', '_major', '_mountain', '_place', 
           '_river', '_state']
BINARIES = [ '_area', '_capital', '_density', '_elevation', '_high_point', 
            '_higher', '_loc', '_longer', '_low_point', '_lower', '_len',
            '_next_to', '_population', '_size', '_traverse']
# Special predicates take in lists of predicates as an argument
# Maps from label to tuple (number of children, child which is an expression)
SPECIAL_PREDICATES = { 
    '_answer': (2, 1),
    '_largest': (2, 1),
    '_smallest': (2, 1),
    '_highest': (2, 1),
    '_lowest': (2, 1),
    '_longest': (2, 1),
    '_shortest': (2, 1),
    '_count': (3, 1),
    '_most': (3, 2),
    '_fewest': (3, 2),
}

def check_is(x, c):
  if x != c: raise ValueError('Expected "%s", got "%s"' % (c, x))
  return x

def check_is_var(x):
  if not (len(x) == 1 and ord(x) >= ord('A') and ord(x) <= ord('Z')):
    raise ValueError('Expected variable, got "%s"' % x)
  return x

def check_is_id(x):
  if x not in IDS:
    raise ValueError('Expected ID, got "%s"' % x)
  return x

def print_ast(node, level=0):
  indent = ' ' * (4 * level)
  print '%s%s' % (indent, node.label)
  for c in node.children:
    print_ast(c, level=level+1)

def parse_funql(y, verbose=0):
  """Parse a FunQL logical form.

  See specification at
  http://www.cs.utexas.edu/~ml/wasp/geo-funql.html

  We parse this into basically the abstract syntax tree.
  Each node in the tree has a label and a list of child nodes.

  A label is either a FunQL predicate, or a special label "_list"
  which denotes that the expression is a list of predicates.
  """
  def recurse(i):
    label = y[i]
    if verbose >= 2:
      print (label, i)
    if label in UNARIES:
      check_is(y[i+1], '(')
      var_node = ASTNode(check_is_var(y[i+2]), [])
      check_is(y[i+3], ')')
      return (ASTNode(label, [var_node]), i+4)
    elif label in BINARIES:
      check_is(y[i+1], '(')
      var_1 = ASTNode(check_is_var(y[i+2]), [])
      check_is(y[i+3], ',')
      var_2 = ASTNode(check_is_var(y[i+4]), [])
      check_is(y[i+5], ')')
      return (ASTNode(label, [var_1, var_2]), i+6)
    elif label in SPECIAL_PREDICATES:
      check_is(y[i+1], '(')
      nargs, expr_ind = SPECIAL_PREDICATES[label]
      next_ind = i+2
      children = []
      for j in range(nargs):
        if j == expr_ind:
          if y[next_ind] == '(':
            # Expression is parenthesized
            next_ind += 1
            sub_children = []
            while True:
              new_child, next_ind = recurse(next_ind)
              sub_children.append(new_child)
              if y[next_ind] == ',': 
                next_ind += 1
              else:
                break
            check_is(y[next_ind], ')')
            next_ind += 1
          else:
            # Expression is a single literal
            new_child, next_ind = recurse(next_ind)
            sub_children = [new_child]
          node = ASTNode('_list', sub_children)
          children.append(node)
        else:
          children.append(ASTNode(check_is_var(y[next_ind]), []))
          next_ind += 1
        if j < nargs - 1:
          check_is(y[next_ind], ',')
          next_ind += 1
      check_is(y[next_ind], ')')
      return (ASTNode(label, children), next_ind+1)
    elif label == '_const':
      # Handle _const separately
      check_is(y[i+1], '(')
      var_node = ASTNode(check_is_var(y[i+2]), [])
      check_is(y[i+3], ',')
      id_type = check_is_id(y[i+4])
      check_is(y[i+5], '(')
      if y[i+6] == "'":
        j = i+7
        toks = []
        while y[j] != "'":
          toks.append(y[j])
          j += 1
        name = ' '.join(toks)
      else:
        name = y[i+6]
        j = i+6
      if id_type == '_cityid':
        check_is(y[j+1], ',')
        name = ' '.join((name, y[j+1], y[j+2]))
        j += 2
      name_node = ASTNode(name, [])
      check_is(y[j+1], ')')
      inner_node = ASTNode(id_type, [name_node])
      check_is(y[j+2], ')')
      return (ASTNode(label, [var_node, inner_node]), j+3)
    elif label == '\\+':
      # Handle not separately
      next_ind = i+1
      if y[i+1] == '(':
        # Expression is parenthesized
        next_ind += 1
        print (next_ind, y[next_ind])
        children = []
        while True:
          new_child, next_ind = recurse(next_ind)
          children.append(new_child)
          if y[next_ind] == ',': 
            next_ind += 1
          else:
            break
        check_is(y[next_ind], ')')
        next_ind += 1
      else:
        # Expression is a single literal
        new_child, next_ind = recurse(next_ind)
        children = [new_child]
      return (ASTNode('_not', children), next_ind)
    else:
      raise ValueError('Unrecognized label "%s"' % label)
  if verbose >= 1:
    print 'Parsing "%s"' % ' '.join(y)
  root, i_final = recurse(0)
  if i_final != len(y):
    raise Exception('Parsing sentence length %d stopped at %d' % (len(y), i_final))
  if verbose >= 1:
    print_ast(root)
  return root


def make_lambda_dcs(y):
  root = parse_funql(y.split(), verbose=1)

def gen_data(in_data):
  return [(x, make_lambda_dcs(y)) for x, y in in_data]

def process(filename):
  print >> sys.stderr, 'Processing %s' % filename
  basename = os.path.basename(filename)
  in_data = read_examples(filename)
  out_data = gen_data(in_data)
  out_path = os.path.join(OUT_DIR, basename)
  sys.exit(0)
  with open(out_path, 'w') as f:
    for x, y in out_data:
      print >>f, '%s\t%s' % (x, y)

def main():
  for filename in sorted(glob.glob(os.path.join(IN_DIR, '*.tsv'))):
    process(filename)


if __name__ == '__main__':
  main()
